
# THIS FILE IS GENERATED FROM CVXPY SETUP.PY
short_version = '1.9.0'
version = '1.9.0'
full_version = '1.9.0'
git_revision = '261539f'
commit_count = '265'
release = True
if not release:
    version = full_version
